//#include <SoftwareSerial.h>
#include "MX3080.h"

#define BAUD_RATE_PC 9600
#define BAUD_RATE_MX3080 9600

#define PIN_RX 3
#define PIN_TX 2

SoftwareSerial mxSerial(PIN_RX, PIN_TX); // RX, TX
void processCommand(String kv);
MX3080 wifi(mxSerial, BAUD_RATE_MX3080, &processCommand);

String periString = "";         // a String to hold incoming data
bool periStringComplete = false;  // whether the string is complete

typedef struct {
  int16_t reset;
  int16_t pairing;
  int16_t sensor0;
  int16_t sensor1;
  int16_t sensor2;
  int16_t sensor3;
  int16_t sensor4;
  int16_t sensor5;
  int16_t sensor6;
  int16_t sensor7;
  int16_t sensor8;
  int16_t sensor9;
  int16_t sensor10;
  int16_t sensor11;
  int16_t sensor12;
  int16_t sensor13;
  int16_t sensor14;
  int16_t sensor15;
  int16_t sensor16;
  int16_t sensor17;
  int16_t sensor18;
  int16_t sensor19;
  int16_t sensor20;
  int16_t sensor21;
  int16_t sensor22;
  int16_t sensor23;
  int16_t sensor24;
  int16_t sensor25;
  int16_t sensor26;
  int16_t sensor27;
  int16_t sensor28;
} sensor_state;

static sensor_state _state = {
  .reset = 0,
  .pairing = 0,
  .sensor0 = 0,
  .sensor1 = 0,
  .sensor2 = 0,
  .sensor3 = 0,
  .sensor4 = 0,
  .sensor5 = 0,
  .sensor6 = 0,
  .sensor7 = 0,
  .sensor8 = 0,
  .sensor9 = 0,
  .sensor10 = 0,
  .sensor11 = 0,
  .sensor12 = 0,
  .sensor13 = 0,
  .sensor14 = 0,
  .sensor15 = 0,
  .sensor16 = 0,
  .sensor17 = 0,
  .sensor18 = 0,
  .sensor19 = 0,
  .sensor20 = 0,
  .sensor21 = 0,
  .sensor22 = 0,
  .sensor23 = 0,
  .sensor24 = 0,
  .sensor25 = 0,
  .sensor26 = 0,
  .sensor27 = 0,
  .sensor28 = 0
};

static char* on = "on";
static char* off = "off";

void setup() {
  Serial.begin(BAUD_RATE_PC);
  periString.reserve(100);

  Serial.println("Edworks!");
  wifi.kick();
}

void loop() {
  wifi.eventLooper();
  if (wifi.shouldUpdateDevice) {
    sendDevstat(&_state);
    wifi.shouldUpdateDevice = false;
  }

  handleSerialEvent();
}

void handleSerialEvent() {
    if (periStringComplete) {
        // 현 peripheral 상태값 update
        updatePeripheralState(periString);
        handleUserAction(&_state);
        if (wifi.shouldSendReport) {
            sendSkstat(&_state);
        } else {
            _state.sensor1 = wifi.getRssi();
        }
        periString = "";
        periStringComplete = false;
        wifi.shouldSendReport = false;
    }

    if (!Serial.available()) return;

    char inChar = (char)Serial.read();
    if (periString.length() == 0 && isspace(inChar) != 0) {
        // 첫글자가 공백이면 skip
        periStringComplete = false;
    } else {
        if (inChar == '\r' || inChar == '\n') {
            periStringComplete = true;
        } else {
            periString += inChar;
        }
    }
}


void sendSkstat(sensor_state *state) {
    static char atCmdBuf[460];
    memset(atCmdBuf, 0x0, 460);
    sprintf_P(atCmdBuf,
        PSTR("power:%s,wifi_strength:%d,temperature:%d,humidity:%d,co2_concentration:%d,voc:%d,fine_dust:%d,fine_dust2:%d,fine_dust3:%d,amount_of_electricity_used:%d,airflow:%02d,power_light1:%s,power_light2:%s,power_light3:%s,power_light4:%s,power_light5:%s,power_light6:%s,power_light_all:%s,dimming:%d,led_bright:%d,colors:%d,direction:%s,speed:%d,canvas:%d,distance:%d,height:%d,intensity_of_illumination:%d,motion_sensor:%02d,open_sense:%02d"),
        (state->sensor0 ? on: off),
        (state->sensor1),
        (state->sensor2),
        (state->sensor3),
        (state->sensor4),
        (state->sensor5),
        (state->sensor6),
        (state->sensor7),
        (state->sensor8),
        (state->sensor9),
        (state->sensor10),
        (state->sensor11 ? on: off),
        (state->sensor12 ? on: off),
        (state->sensor13 ? on: off),
        (state->sensor14 ? on: off),
        (state->sensor15 ? on: off),
        (state->sensor16 ? on: off),
        (state->sensor17 ? on: off),
        (state->sensor18),
        (state->sensor19),
        (state->sensor20),
        (state->sensor21 == 0 ? F("up"): state->sensor21 == 1 ? F("down"): state->sensor21 == 2 ? F("right"): state->sensor21 == 3 ? F("left"): state->sensor21 == 4 ? F("front"): F("back")),
        (state->sensor22),
        (state->sensor23),
        (state->sensor24),
        (state->sensor25),
        (state->sensor26),
        (state->sensor27),
        (state->sensor28)
    );
    wifi.skStat(atCmdBuf);
}

void sendDevstat(sensor_state *state) {
    static char charBuf[64];
    memset(charBuf, 0x0, 64);
    sprintf_P(charBuf,
        PSTR("#%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x"),
        state->reset, state->pairing,
        state->sensor0,
        state->sensor1,
        state->sensor2,
        state->sensor3,
        state->sensor4,
        state->sensor5,
        state->sensor6,
        state->sensor7,
        state->sensor8,
        state->sensor9,
        state->sensor10,
        state->sensor11,
        state->sensor12,
        state->sensor13,
        state->sensor14,
        state->sensor15,
        state->sensor16,
        state->sensor17,
        state->sensor18,
        state->sensor19,
        state->sensor20,
        state->sensor21,
        state->sensor22,
        state->sensor23,
        state->sensor24,
        state->sensor25,
        state->sensor26,
        state->sensor27,
        state->sensor28);
    Serial.println(charBuf);
}

void updatePeripheralState(String periString) {

 /* periString : hexa byte string */
  int8_t periCharBuf[4*31 + 1] = { 0 };  // 각 16bit -> string 0000 -> 4byte, null ended
  periString.toCharArray(periCharBuf, 4*31+1);

  int16_t* periData = wifi.hex2int16(periCharBuf);

  wifi.shouldSendReport = wifi.shouldSendReport
                    | (_state.sensor0 != periData[2])
                    //| (_state.sensor1 != periData[3]) -- RSSI
                    | (_state.sensor2 != periData[4])
                    | (_state.sensor3 != periData[5])
                    | (_state.sensor4 != periData[6])
                    | (_state.sensor5 != periData[7])
                    | (_state.sensor6 != periData[8])
                    | (_state.sensor7 != periData[9])
                    | (_state.sensor8 != periData[10])
                    | (_state.sensor9 != periData[11])
                    | (_state.sensor10 != periData[12])
                    | (_state.sensor11 != periData[13])
                    | (_state.sensor12 != periData[14])
                    | (_state.sensor13 != periData[15])
                    | (_state.sensor14 != periData[16])
                    | (_state.sensor15 != periData[17])
                    | (_state.sensor16 != periData[18])
                    | (_state.sensor17 != periData[19])
                    | (_state.sensor18 != periData[20])
                    | (_state.sensor19 != periData[21])
                    | (_state.sensor20 != periData[22])
                    | (_state.sensor21 != periData[23])
                    | (_state.sensor22 != periData[24])
                    | (_state.sensor23 != periData[25])
                    | (_state.sensor24 != periData[26])
                    | (_state.sensor25 != periData[27])
                    | (_state.sensor26 != periData[28])
                    | (_state.sensor27 != periData[29])
                    | (_state.sensor28 != periData[30]);

  _state.reset = periData[0];
  _state.pairing = periData[1];
  _state.sensor0 = periData[2];
  // _state.sensor1 = periData[3]; RSSI
  _state.sensor2 = periData[4];
  _state.sensor3 = periData[5];
  _state.sensor4 = periData[6];
  _state.sensor5 = periData[7];
  _state.sensor6 = periData[8];
  _state.sensor7 = periData[9];
  _state.sensor8 = periData[10];
  _state.sensor9 = periData[11];
  _state.sensor10 = periData[12];
  _state.sensor11 = periData[13];
  _state.sensor12 = periData[14];
  _state.sensor13 = periData[15];
  _state.sensor14 = periData[16];
  _state.sensor15 = periData[17];
  _state.sensor16 = periData[18];
  _state.sensor17 = periData[19];
  _state.sensor18 = periData[20];
  _state.sensor19 = periData[21];
  _state.sensor20 = periData[22];
  _state.sensor21 = periData[23];
  _state.sensor22 = periData[24];
  _state.sensor23 = periData[25];
  _state.sensor24 = periData[26];
  _state.sensor25 = periData[27];
  _state.sensor26 = periData[28];
  _state.sensor27 = periData[29];
  _state.sensor28 = periData[30];

  if (periData) free(periData);
}

void handleUserAction(sensor_state *state) {
  if (state->reset == 1) {
    wifi.restart();
  }
  if (state->pairing == 1) {
      wifi.skPair();
  }
}

void processCommand(String kv) {
  int idx = kv.indexOf(":");
  if (idx < 1) {
    return;
  }

  String command = kv.substring(0, idx);
  String value = kv.substring(idx+1);
  value.trim();

  if (command.equals(F("power"))) {
    _state.sensor0 = value.equals(on) ? 1: 0;
  } else if (command.equals(F("power_light1"))) {
    _state.sensor11 = value.equals(on);
  } else if (command.equals(F("power_light2"))) {
    _state.sensor12 = value.equals(on);
  } else if (command.equals(F("power_light3"))) {
    _state.sensor13 = value.equals(on);
  } else if (command.equals(F("power_light4"))) {
    _state.sensor14 = value.equals(on);
  } else if (command.equals(F("power_light5"))) {
    _state.sensor15 = value.equals(on);
  } else if (command.equals(F("power_light6"))) {
    _state.sensor16 = value.equals(on);
  } else if (command.equals(F("power_light_all"))) {
    _state.sensor17 = value.equals(on);
  } else if (command.equals(F("airflow"))) {
    _state.sensor10 = value.equals(F("01")) ? 1: value.equals(F("02")) ? 2: value.equals(F("03")) ? 3: value.equals(F("04")) ? 4: value.equals(F("05")) ? 5: value.equals(F("06")) ? 6: value.equals(F("07")) ? 7: value.equals(F("08")) ? 8: 7;
  } else if (command.equals(F("direction"))) {
    _state.sensor21 = value.equals(F("up")) ? 0: value.equals(F("up")) ? 0: value.equals(F("down")) ? 1: value.equals(F("left")) ? 2: value.equals(F("right")) ? 3: value.equals(F("front")) ? 4: value.equals(F("back")) ? 5: 4;
  } else if (command.equals(F("dimming"))) {
    _state.sensor18 = value.toInt();
  } else if (command.equals(F("led_bright"))) {
    _state.sensor19 = value.toInt();
  } else if (command.equals(F("colors"))) {
    _state.sensor20 = value.toInt();
  } else if (command.equals(F("speed"))) {
    _state.sensor22 = value.toInt();
  }
}
